@@ -24,11 +24,6 @@ extern int cr_check(void);
 extern int cr_exec(int pid, char **opts);
 extern int cr_dedup(void);
 
-extern int restrict_uid(unsigned int uid, unsigned int gid);
-struct proc_status_creds;
-extern bool may_dump(struct proc_status_creds *);
-struct _CredsEntry;
-extern bool may_restore(struct _CredsEntry *);
-extern bool cr_user_is_root(void);
+extern int check_add_feature(char *arg);
 
 #endif /* __CR_CRTOOLS_H__ */
