@@ -17,7 +17,7 @@
 #include "string.h"
 #include "sockets.h"
 #include "cr_options.h"
-
+#include "bfd.h"
 #include "protobuf.h"
 
 /*
@@ -460,14 +460,14 @@ static void pb_show_msg(const void *msg, pb_pr_ctl_t *ctl)
 	}
 }
 
-static inline void pb_no_payload(int fd, void *obj) { }
+static inline void pb_no_payload(struct cr_img *i, void *obj) { }
 
-void do_pb_show_plain(int fd, int type, int single_entry,
-		void (*payload_hadler)(int fd, void *obj),
+void do_pb_show_plain(struct cr_img *img, int type, int single_entry,
+		void (*payload_hadler)(struct cr_img *, void *obj),
 		const char *pretty_fmt)
 {
 	pb_pr_ctl_t ctl = {NULL, single_entry, pretty_fmt};
-	void (*handle_payload)(int fd, void *obj);
+	void (*handle_payload)(struct cr_img *, void *obj);
 
 	if (!cr_pb_descs[type].pb_desc) {
 		pr_err("Wrong object requested %d\n", type);
@@ -479,12 +479,12 @@ void do_pb_show_plain(int fd, int type, int single_entry,
 	while (1) {
 		void *obj;
 
-		if (pb_read_one_eof(fd, &obj, type) <= 0)
+		if (pb_read_one_eof(img, &obj, type) <= 0)
 			break;
 
 		ctl.arg = (void *)cr_pb_descs[type].pb_desc;
 		pb_show_msg(obj, &ctl);
-		handle_payload(fd, obj);
+		handle_payload(img, obj);
 		cr_pb_descs[type].free(obj, NULL);
 		if (single_entry)
 			break;
@@ -492,8 +492,9 @@ void do_pb_show_plain(int fd, int type, int single_entry,
 	}
 }
 
-static char *image_name(int fd)
+static char *image_name(struct cr_img *img)
 {
+	int fd = img->_x.fd;
 	static char image_path[PATH_MAX];
 
 	if (read_fd_link(fd, image_path, sizeof(image_path)) > 0)
@@ -512,7 +513,7 @@ static char *image_name(int fd)
  * Don't forget to free memory granted to unpacked object in calling code if needed
  */
 
-int do_pb_read_one(int fd, void **pobj, int type, bool eof)
+int do_pb_read_one(struct cr_img *img, void **pobj, int type, bool eof)
 {
 	u8 local[PB_PKOBJ_LOCAL_SIZE];
 	void *buf = (void *)&local;
@@ -521,25 +522,25 @@ int do_pb_read_one(int fd, void **pobj, int type, bool eof)
 
 	if (!cr_pb_descs[type].pb_desc) {
 		pr_err("Wrong object requested %d on %s\n",
-			type, image_name(fd));
+			type, image_name(img));
 		return -1;
 	}
 
 	*pobj = NULL;
 
-	ret = read(fd, &size, sizeof(size));
+	ret = bread(&img->_x, &size, sizeof(size));
 	if (ret == 0) {
 		if (eof) {
 			return 0;
 		} else {
 			pr_err("Unexpected EOF on %s\n",
-			       image_name(fd));
+			       image_name(img));
 			return -1;
 		}
 	} else if (ret < sizeof(size)) {
 		pr_perror("Read %d bytes while %d expected on %s",
 			  ret, (int)sizeof(size),
-			  image_name(fd));
+			  image_name(img));
 		return -1;
 	}
 
@@ -550,14 +551,14 @@ int do_pb_read_one(int fd, void **pobj, int type, bool eof)
 			goto err;
 	}
 
-	ret = read(fd, buf, size);
+	ret = bread(&img->_x, buf, size);
 	if (ret < 0) {
 		pr_perror("Can't read %d bytes from file %s",
-			  size, image_name(fd));
+			  size, image_name(img));
 		goto err;
 	} else if (ret != size) {
 		pr_perror("Read %d bytes while %d expected from %s",
-			  ret, size, image_name(fd));
+			  ret, size, image_name(img));
 		ret = -1;
 		goto err;
 	}
@@ -566,7 +567,7 @@ int do_pb_read_one(int fd, void **pobj, int type, bool eof)
 	if (!*pobj) {
 		ret = -1;
 		pr_err("Failed unpacking object %p from %s\n",
-		       pobj, image_name(fd));
+		       pobj, image_name(img));
 		goto err;
 	}
 
@@ -586,7 +587,7 @@ err:
  *  0 on success
  * -1 on error
  */
-int pb_write_one(int fd, void *obj, int type)
+int pb_write_one(struct cr_img *img, void *obj, int type)
 {
 	u8 local[PB_PKOBJ_LOCAL_SIZE];
 	void *buf = (void *)&local;
@@ -617,7 +618,7 @@ int pb_write_one(int fd, void *obj, int type)
 	iov[1].iov_base = buf;
 	iov[1].iov_len = size;
 
-	ret = writev(fd, iov, 2);
+	ret = bwritev(&img->_x, iov, 2);
 	if (ret != size + sizeof(size)) {
 		pr_perror("Can't write %d bytes", (int)(size + sizeof(size)));
 		goto err;
@@ -633,16 +634,17 @@ err:
 int collect_image(struct collect_image_info *cinfo)
 {
 	bool optional = !!(cinfo->flags & COLLECT_OPTIONAL);
-	int fd, ret;
+	int ret;
+	struct cr_img *img;
 	void *(*o_alloc)(size_t size) = malloc;
 	void (*o_free)(void *ptr) = free;
 
 	pr_info("Collecting %d/%d (flags %x)\n",
 			cinfo->fd_type, cinfo->pb_type, cinfo->flags);
 
-	fd = open_image(cinfo->fd_type, O_RSTR | (optional ? O_OPT : 0));
-	if (fd < 0) {
-		if (optional && fd == -ENOENT)
+	img = open_image(cinfo->fd_type, O_RSTR | (optional ? O_OPT : 0));
+	if (!img) {
+		if (optional && errno == ENOENT)
 			return 0;
 		else
 			return -1;
@@ -666,7 +668,7 @@ int collect_image(struct collect_image_info *cinfo)
 		} else
 			obj = NULL;
 
-		ret = pb_read_one_eof(fd, &msg, cinfo->pb_type);
+		ret = pb_read_one_eof(img, &msg, cinfo->pb_type);
 		if (ret <= 0) {
 			o_free(obj);
 			break;
@@ -683,7 +685,7 @@ int collect_image(struct collect_image_info *cinfo)
 			cr_pb_descs[cinfo->pb_type].free(msg, NULL);
 	}
 
-	close(fd);
+	close_image(img);
 	pr_debug(" `- ... done\n");
 	return ret;
 }
