@@ -20,20 +20,32 @@ sigset_t mask;
 #define MAX_TIMER_DISPLACEMENT	10
 #define NO_PERIODIC
 
+#ifndef CLOCK_MONOTONIC_COARSE
+# define CLOCK_MONOTONIC_COARSE	6
+#endif
+
+#ifndef CLOCK_BOOTTIME
+# define CLOCK_BOOTTIME		7
+#endif
+
 #ifndef NO_PERIODIC
 static void realtime_periodic_handler(int sig, siginfo_t *si, void *uc);
 static void monotonic_periodic_handler(int sig, siginfo_t *si, void *uc);
+static void boottime_periodic_handler(int sig, siginfo_t *si, void *uc);
 #endif
 static void realtime_oneshot_handler(int sig, siginfo_t *si, void *uc);
 static void monotonic_oneshot_handler(int sig, siginfo_t *si, void *uc);
+static void boottime_oneshot_handler(int sig, siginfo_t *si, void *uc);
 
 enum {
 #ifndef NO_PERIODIC
 	REALTIME_PERIODIC_INFO,
 	MONOTONIC_PERIODIC_INFO,
+	BOOTTIME_PERIODIC_INFO,
 #endif
 	REALTIME_ONESHOT_INFO,
 	MONOTONIC_ONESHOT_INFO,
+	BOOTTIME_ONESHOT_INFO,
 };
 
 static struct posix_timers_info {
@@ -55,11 +67,15 @@ static struct posix_timers_info {
 				realtime_periodic_handler, SIGALRM, 0, 1},
 	[MONOTONIC_PERIODIC_INFO] = {CLOCK_MONOTONIC, "MONOTONIC (periodic)",
 				monotonic_periodic_handler, SIGINT, 0, 3},
+	[BOOTTIME_PERIODIC_INFO] = {CLOCK_BOOTTIME, "BOOTTIME (periodic)",
+				boottime_periodic_handler, SIGWINCH, 0, 3},
 #endif
 	[REALTIME_ONESHOT_INFO] = {CLOCK_REALTIME, "REALTIME (oneshot)",
 				realtime_oneshot_handler, SIGUSR1, 1, INT_MAX},
 	[MONOTONIC_ONESHOT_INFO] = {CLOCK_MONOTONIC, "MONOTONIC (oneshot)",
 				monotonic_oneshot_handler, SIGUSR2, 1, INT_MAX},
+	[BOOTTIME_ONESHOT_INFO] = {CLOCK_BOOTTIME, "BOOTTIME (oneshot)",
+				boottime_oneshot_handler, SIGPROF, 1, INT_MAX },
 	{ }
 };
 
@@ -132,7 +148,7 @@ static int check_handler_status(struct posix_timers_info *info,
 	return 0;
 }
 
-static int check_timers(int delta)
+static int check_timers(int delta, struct timespec *sleep_start, struct timespec *sleep_end)
 {
 	struct posix_timers_info *info = posix_timers;
 	int ms_passed;
@@ -156,6 +172,16 @@ static int check_timers(int delta)
 			return -errno;
 		}
 
+		/*
+		 * Adjust with @total_sleep_time if needed.
+		 */
+		if (info->clock == CLOCK_BOOTTIME) {
+			info->start.tv_sec	-= sleep_start->tv_sec;
+			info->start.tv_nsec	-= sleep_start->tv_nsec;
+			info->end.tv_sec	-= sleep_end->tv_sec;
+			info->end.tv_nsec	-= sleep_end->tv_nsec;
+		}
+
 		ms_passed = (info->end.tv_sec - info->start.tv_sec) * 1000 +
 			(info->end.tv_nsec - info->start.tv_nsec) / (1000 * 1000);
 
@@ -171,6 +197,9 @@ static void generic_handler(struct posix_timers_info *info,
 {
 	int overrun;
 
+	if (info == NULL)
+		info = &posix_timers[MONOTONIC_ONESHOT_INFO];
+
 	if (info != real) {
 		real->handler_status |= WRONG_SI_PTR;
 		return;
@@ -193,6 +222,12 @@ static void monotonic_periodic_handler(int sig, siginfo_t *si, void *uc)
 	generic_handler(si->si_value.sival_ptr,
 			&posix_timers[MONOTONIC_PERIODIC_INFO], sig);
 }
+
+static void boottime_periodic_handler(int sig, siginfo_t *si, void *uc)
+{
+	generic_handler(si->si_value.sival_ptr,
+			&posix_timers[BOOTTIME_PERIODIC_INFO], sig);
+}
 #endif
 
 static void monotonic_oneshot_handler(int sig, siginfo_t *si, void *uc)
@@ -201,6 +236,12 @@ static void monotonic_oneshot_handler(int sig, siginfo_t *si, void *uc)
 			&posix_timers[MONOTONIC_ONESHOT_INFO], sig);
 }
 
+static void boottime_oneshot_handler(int sig, siginfo_t *si, void *uc)
+{
+	generic_handler(si->si_value.sival_ptr,
+			&posix_timers[BOOTTIME_ONESHOT_INFO], sig);
+}
+
 #ifndef NO_PERIODIC
 static void realtime_periodic_handler(int sig, siginfo_t *si, void *uc)
 {
@@ -262,7 +303,10 @@ static int setup_timers(void)
 
 		sev.sigev_notify = SIGEV_SIGNAL;
 		sev.sigev_signo = info->sig;
-		sev.sigev_value.sival_ptr = info;
+		if (&posix_timers[MONOTONIC_ONESHOT_INFO] == info)
+			sev.sigev_value.sival_ptr = NULL;
+		else
+			sev.sigev_value.sival_ptr = info;
 
 		if (timer_create(info->clock, &sev, &info->timerid) == -1) {
 			err("Can't create timer\n");
@@ -291,8 +335,43 @@ static int setup_timers(void)
 	return 0;
 }
 
+/*
+ * Figure out @total_sleep_time, ie time the system was in hardware
+ * suspend mode, will need this value to exclude from boottime clock
+ * testing.
+ */
+static int get_total_sleep_time(struct timespec *tv, char *type)
+{
+	struct timespec boottime_coarse;
+	struct timespec boottime;
+
+	if (clock_gettime(CLOCK_BOOTTIME, &boottime) == -1) {
+		err("Can't get CLOCK_BOOTTIME %s time\n", type);
+		return -errno;
+	}
+
+	if (clock_gettime(CLOCK_MONOTONIC_COARSE, &boottime_coarse) == -1) {
+		err("Can't get CLOCK_MONOTONIC_COARSE %s time\n", type);
+		return -errno;
+	}
+
+	tv->tv_sec = boottime.tv_sec - boottime_coarse.tv_sec;
+	tv->tv_nsec = boottime.tv_nsec - boottime_coarse.tv_nsec;
+
+	test_msg("(%6s) boottime %lu "
+		 "boottime-coarse %lu "
+		 "total_sleep_time %lu\n",
+		 type,
+		 (long)boottime.tv_sec,
+		 (long)boottime_coarse.tv_sec,
+		 (long)tv->tv_sec);
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
+	struct timespec sleep_start, sleep_end;
 	struct timespec start, end;
 	int err;
 
@@ -307,11 +386,19 @@ int main(int argc, char **argv)
 	test_daemon();
 
 	clock_gettime(CLOCK_REALTIME, &start);
-	test_waitsig();
-	clock_gettime(CLOCK_REALTIME, &end);
+	err = get_total_sleep_time(&sleep_start, "start");
+	if (err)
+		return err;
 
+	test_waitsig();
+
+	clock_gettime(CLOCK_REALTIME, &end);
+	err = get_total_sleep_time(&sleep_end, "end");
+	if (err)
+		return err;
 	err = check_timers((end.tv_sec - start.tv_sec) * 1000 +
-				(end.tv_nsec - start.tv_nsec) / 1000000);
+			   (end.tv_nsec - start.tv_nsec) / 1000000,
+			   &sleep_start, &sleep_end);
 	if (err)
 		return err;
 
